/* 
 * $Id: shuffle.c,v 1.7 1997/03/06 18:46:51 drz Exp drz $
 * by Tim Newsome (drz@cybercomm.net)
 * This program shuffles its command line arguments and prints them out on a
 * single line. 
 * $Log: shuffle.c,v $
 * Revision 1.7  1997/03/06 18:46:51  drz
 * Fixed where we were writing outside our memory...
 *
 * Revision 1.6  1997/03/06 00:20:44  drz
 * Now includes -n.
 *
 * Revision 1.5  1996/08/02 00:27:32  drz
 * Added/moved comments, add a free() to fix a potential bug. Played
 * with error codes.
 *
 * Revision 1.4  1996/08/01 23:16:59  drz
 * Did I change anything?
 *
 * Revision 1.3  1996/07/21 15:21:48  drz
 * Added the -v,--version options.
 *
 * Revision 1.2  1996/07/21 01:28:30  drz
 * Added the "Not enough memory." error message.
 *
 * Revision 1.1  1996/07/20 22:35:28  drz
 * Now it works. :-) There are 2 blocks of memory. One contains the entire
 * file, the other is an index to the file, containing offsets from the
 * start of the file.
 *
 * Revision 1.0  1996/06/27 01:30:37  drz
 * Initial revision
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <getopt.h>

#include "shuffle.h"

static int commandline(int n, int argc, char *argv[]);
static int input(int n);

int main(int argc, char *argv[])
{
	int ret;
	int c;
	int n = -1;
	int option_index;
	static char version[] = "Shuffle version 1.2\n";

	/*
	 * Parse any commandline arguments.
	 */
	static struct option long_options[] =
	{
		{"version", no_argument, 0, 0},
		{"number", required_argument, 0, 0},
		{0, 0, 0, 0}};
	option_index = 0;

	while ((c = getopt_long(argc, argv, "n:v", long_options, &option_index))
				 != -1) {
		switch (c) {
		case 0:										/* A long option. */
			if (!strcmp("version", long_options[option_index].name)) {
				fprintf(stderr, version);
				return (0);

			} else if (!strcmp("number", long_options[option_index].name)) {
				n = atoi(optarg);
			}
			break;

		case 'n':
			n = atoi(optarg);
			break;

		case 'v':
			fprintf(stderr, version);
			return (0);

		default:
			return (1);
		}
	}

	/*
	 * Hide the arguments from the actual shuffling program.
	 */
	argc -= optind - 1;
	argv += optind - 1;

	/*
	 * Seed the random number generator. 
	 */
	srandom((int) time(0));

	/*
	 * If there are no arguments. 
	 */
	if (argc == 1) {
		ret = input(n);
	} else {
		/*
		 * Shuffle the commandline, not including the name of the program. 
		 */
		ret = commandline(n, --argc, ++argv);
	}

	switch (ret) {
	case SE_NOMEM:
		fprintf(stderr, "%s: %s\n",
					rindex(argv[0], '/') == NULL ? argv[0] : rindex(argv[0], '/') + 1,
						SE_NOMEM_STR);
		break;
	}

	return (ret);
}

/* 
 * This function is called to shuffle the commandline, which we already
 * have in the form of an index to an array of strings.
 */
int commandline(int n, int argc, char *argv[])
{
	int r;

	while (argc && n--) {
		r = random() % argc;				/* Pick a random argument. */
		fputs(argv[r], stdout);			/* Print it. */
		putchar(' ');								/* Add a space. */
		/*
		 * Move the last element from the list onto the argument we just
		 * printed and decrement argc. 
		 */
		argv[r] = argv[--argc];
	}
	return (SE_NOERR);
}

/* 
 * Shuffle stdin.
 */
int input(int n)
{
	long *index_start;						/* Where our index starts. */
	long *new_index;							/* Temporary variable. */
	long index_size;
	long index_index;							/* Which entry we're currently looking at. */
	char *data_start;							/* Where our data starts. */
	char *new_data;								/* Temporary variable. */
	long data_size;
	long data_index;							/* Which entry we're currently looking at. */
	char c;
	long r;
	char *p;

	/*
	 * First, get some memory to start of with.
	 */
	data_size = DATA_INCREMENT;
	data_start = malloc(data_size * sizeof(char));
	if (data_start == NULL) {
		return (SE_NOMEM);
	}
	index_size = INDEX_INCREMENT;
	index_start = malloc(index_size * sizeof(char *));
	if (index_start == NULL) {
		free(data_start);
		return (SE_NOMEM);
	}
	/*
	 * Read all lines into memory and index it while we're doing it.
	 */
	data_index = 0;
	index_start[0] = 0;
	index_index = 1;

	while ((c = getchar()) != EOF) {
		if (data_index >= data_size) {
			/*
			 * Get more data memory. 
			 */
			data_size += DATA_INCREMENT;
			new_data = realloc(data_start, data_size * sizeof(char));
			if (new_data == NULL) {
				/*
				 * Free index_start first because there might be a fatal problem
				 * freeing data_start.
				 */
				free(index_start);
				free(data_start);
				return (SE_NOMEM);
			}
			data_start = new_data;
		}
		data_start[data_index++] = c;

		if (c == SEPARATOR) {
			/*
			 * Add an entry to the index. 
			 */
			if (index_index >= index_size) {
				/*
				 * Add more index memory. 
				 */
				index_size += INDEX_INCREMENT;
				new_index = realloc(index_start, index_size * sizeof(char *));
				if (new_index == NULL) {
					/*
					 * Free data_start first because there might be a fatal problem
					 * freeing index_start.
					 */
					free(data_start);
					free(index_start);
					return (SE_NOMEM);
				}
				index_start = new_index;
			}
			index_start[index_index++] = data_index;
		}
	}															/* EOF */

	data_start[data_index] = SEPARATOR;
	index_size = index_index - 1;

	/*
	 * Shuffle the stuff and print it as we go.
	 */

	while (index_size && n--) {
		r = random() % index_size;	/* Pick a random argument. */
		/*
		 * Print it.
		 */
		p = &data_start[index_start[r]];
		while (*p != SEPARATOR) {
			putchar(*p++);
		}
		putchar(SEPARATOR);
		/*
		 * Move the last element from the list onto the argument we just
		 * printed. 
		 */
		index_start[r] = index_start[--index_size];
	}

	/*
	 * Free up the memory again.
	 */
	free(data_start);
	free(index_start);

	return (SE_NOERR);
}
